/*
 * Copyright (c) 2002, Stuart Walsh
 * All rights reserved.
 *
 * This file is licensed under the terms of the GPL license.  More details can
 * be found from the file COPYING.
 *
 * Furthermore, this file contains code from iproute copyright  Alexey Kuznetsov,
 * <kuznet@ms2.inr.ac.ru>
 *
 * and BSD ifconfig which is Copyright (c) 1983, 1993 
 * The Regents of the University of California
 *
 * */

#include "ac-config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifndef LINUX
#include <net/if.h>
#include <net/if_var.h>
#include <sys/sockio.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet6/nd6.h>   /* Define ND6_INFINITE_LIFETIME */
#include <ifaddrs.h>
#else
#include <linux/sockios.h>
#include <linux/ip.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/if_tunnel.h>
#include "libnetlink.h"
#endif
#include <netdb.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>

#define ifra_dstaddr ifra_broadaddr

#ifdef LINUX
struct idxmap
{
    struct idxmap * next;
    int     index;
    int     type;
    int     alen;
    unsigned    flags;
    unsigned char   addr[8];
    char        name[16];
};

static struct idxmap *idxmap[16];

int ll_remember_index(struct sockaddr_nl *who, struct nlmsghdr *n, void *arg)
{
    int h;
    struct ifinfomsg *ifi = NLMSG_DATA(n);
    struct idxmap *im, **imp;
    struct rtattr *tb[IFLA_MAX+1];

    if (n->nlmsg_type != RTM_NEWLINK)
        return 0;

    if (n->nlmsg_len < NLMSG_LENGTH(sizeof(ifi)))
        return -1;


    memset(tb, 0, sizeof(tb));
    parse_rtattr(tb, IFLA_MAX, IFLA_RTA(ifi), IFLA_PAYLOAD(n));
    if (tb[IFLA_IFNAME] == NULL)
        return 0;

    h = ifi->ifi_index&0xF;

    for (imp=&idxmap[h]; (im=*imp)!=NULL; imp = &im->next)     
        if (im->index == ifi->ifi_index)
            break;

    if (im == NULL) {
        im = malloc(sizeof(*im));
        if (im == NULL)
            return 0;
        im->next = *imp;
        im->index = ifi->ifi_index;
        *imp = im;
    }
    im->type = ifi->ifi_type;
    im->flags = ifi->ifi_flags;
    if (tb[IFLA_ADDRESS]) {
        int alen;
        im->alen = alen = RTA_PAYLOAD(tb[IFLA_ADDRESS]);
        if (alen > sizeof(im->addr))
            alen = sizeof(im->addr);
        memcpy(im->addr, RTA_DATA(tb[IFLA_ADDRESS]), alen);
    } else {
        im->alen = 0;
        memset(im->addr, 0, sizeof(im->addr));
    }
    strcpy(im->name, RTA_DATA(tb[IFLA_IFNAME]));
    return 0;
}


int ll_name_to_index(char *name)
{
    static char ncache[16];
    static int icache;
    struct idxmap *im;
    int i; 
    
    if (name == NULL)
        return 0;
    if (icache && strcmp(name, ncache) == 0)
        return icache;
    for (i=0; i<16; i++) {
        for (im = idxmap[i]; im; im = im->next) {               
            if (strcmp(im->name, name) == 0) {
                icache = im->index;
                strcpy(ncache, name);
                return im->index;
            }
        }
    }
    return 0;
}

int ll_init_map(struct rtnl_handle *rth)
{                                                             
    if (rtnl_wilddump_request(rth, AF_UNSPEC, RTM_GETLINK) < 0) {
        perror("Cannot send dump request");
        exit(1);
    }

    if (rtnl_dump_filter(rth, ll_remember_index, &idxmap, NULL, NULL) < 0) {
        fprintf(stderr, "Dump terminated\n");
        exit(1);
    }
    return 0;
}

#endif
int create_tunnel(char *interface, char *src, char *dst)
{
#ifndef LINUX
    int s;
    int type = SIOCIFCREATE;
    struct ifaliasreq addreq;
    struct addrinfo *srcres, *dstres;
    struct ifreq ifr;

    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s == -1)
    {
        perror("ERROR: Failed kernel socket()");
        return -1;
    }

    memset(&ifr, 0, sizeof(ifr));
    (void) strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name)-1);
    if (ioctl(s, type, &ifr) < 0)
    {
        fprintf(stderr, "%s: %s\n", "SIOCIFCREATE", strerror(errno));
        return -1;
    }

    if (getaddrinfo(src, NULL, NULL, &srcres) != 0)
    {
        fprintf(stderr, "Error in source address");
        close(s);
        return -1;
    }
    if (getaddrinfo(dst, NULL, NULL, &dstres) != 0)
    {
         fprintf(stderr, "Error in destination address");
         close(s);
         return -1;
    }

    memset(&addreq, 0, sizeof(addreq));
    strncpy(addreq.ifra_name, interface, IFNAMSIZ);
    memcpy(&addreq.ifra_addr, srcres->ai_addr, srcres->ai_addr->sa_len);
    memcpy(&addreq.ifra_dstaddr, dstres->ai_addr, dstres->ai_addr->sa_len);

    if (ioctl(s, SIOCSIFPHYADDR, &addreq) < 0)
    {
        perror("SIOCSIFPHYADDR");
        close(s);
        return -1;
    }
    close(s);
    return 0;
#else
    struct ip_tunnel_parm p;
    struct ifreq ifr;
    int fd;
    int err;
    int flags = 0;
    int mask = 0;

    memset(&p, 0, sizeof(p));
    memset(&ifr, 0, sizeof(ifr));
    p.iph.version = 4;
    p.iph.ihl = 5;
    p.iph.protocol = IPPROTO_IPV6;
    p.iph.saddr = inet_addr(src);
    p.iph.daddr = inet_addr(dst); 
    p.iph.ttl = 64;
    strncpy(p.name, interface, IFNAMSIZ);

    strncpy(ifr.ifr_name, "sit0", IFNAMSIZ);
    ifr.ifr_ifru.ifru_data = (void*)&p;
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    err = ioctl(fd, SIOCADDTUNNEL, &ifr);
    if (err)
    {
        perror("ioctl");
	close(fd);
	return err;
    }
    mask |= IFF_UP;
    flags |= IFF_UP;
		 
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, interface, IFNAMSIZ);
    ifr.ifr_flags &= ~mask;
    ifr.ifr_flags |= mask&flags;
    err = ioctl(fd, SIOCSIFFLAGS, &ifr);
    if (err)
        perror("ioctl");
	
    close(fd);
    return err;
#endif
}

int add_endpoints(char *name, char *src6, char *dst6)
{
#ifndef LINUX
    struct addrinfo hints, *res, *res2;
    struct sockaddr_in6 pref;
    struct in6_ifreq req;
    struct in6_aliasreq  addreq = 
    { { 0 },
        { 0 },
        { 0 },
        { 0 },
        0,
        {0,0, ND6_INFINITE_LIFETIME, ND6_INFINITE_LIFETIME } };
          
    int error = -1, error2 = -1;
    int s;
            
    memset(&req, 0, sizeof(req));
    memset(&pref, 0, sizeof(pref));
    
    bzero(&hints, sizeof(struct addrinfo));
    hints.ai_family = AF_INET6;
    error = getaddrinfo(src6, NULL, &hints, &res);
    error2 = getaddrinfo(dst6, NULL, &hints, &res2);

    if (error != 0) 
    {
        if (inet_pton(AF_INET6, src6, &addreq.ifra_addr) != 1)
        {
            fprintf(stderr, "ERROR: Bad source address\n");
            return -1;
        }
    }
    else
        bcopy(res->ai_addr, &addreq.ifra_addr, res->ai_addrlen);

    if(res)
        freeaddrinfo(res);
    if(error2 != 0)
    {
        if (inet_pton(AF_INET6, dst6, &addreq.ifra_dstaddr) != 1)
        {
            fprintf(stderr, "ERROR: Bad destination address\n");
            return -1;
        }
    }
    else
        bcopy(res2->ai_addr, &addreq.ifra_dstaddr, res2->ai_addrlen);
    
    if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
    {
        perror("kernel socket");
        return -1;
    }
    pref.sin6_len = sizeof(pref);
    memset(&pref.sin6_addr, 0xff, sizeof(struct in6_addr));

    bcopy(&pref, &addreq.ifra_prefixmask, sizeof(pref));
    strncpy(addreq.ifra_name, name, sizeof(addreq.ifra_name));
    strncpy(req.ifr_name, name, sizeof(req.ifr_name));
    /* Clear the old address */
    if(ioctl(s, SIOCDIFADDR_IN6, &req) < 0)
    {
        if(errno != EADDRNOTAVAIL)
        {
            perror("SIOCDIFADDR_IN6");
            return -1;
        }
    }
    
    if(ioctl(s, SIOCAIFADDR_IN6, &addreq) < 0)
    {
        perror("SIOCAIFADDR_IN6");
        return -1;
    }
#else
    struct rtnl_handle rth;
    struct {
        struct nlmsghdr     n;
        struct ifaddrmsg    ifa;
        char            buf[256];
    } req;
    struct in6_addr data;

    
    memset(&req, 0, sizeof(req));
    memset(&data, 0, sizeof(data));
    req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
    req.n.nlmsg_flags = NLM_F_REQUEST;
    req.n.nlmsg_type = RTM_NEWADDR;
    req.ifa.ifa_family = AF_INET6;

    inet_pton(AF_INET6, src6, &data);
    req.ifa.ifa_prefixlen = 128;
    addattr_l(&req.n, sizeof(req), IFA_LOCAL, &data, sizeof(data));

    if (rtnl_open(&rth, 0) < 0)
        exit(1);

    ll_init_map(&rth);
    if ((req.ifa.ifa_index = ll_name_to_index(name)) == 0) 
    {
        fprintf(stderr, "Cannot find device \"%s\"\n", name);
        return -1;
    }
                        
    if (rtnl_talk(&rth, &req.n, 0, 0, NULL, NULL, NULL) < 0)
        exit(2);
#endif
    return 0;
}

int update_tunnel(char *name, char *src, char *dst)
{
#ifndef LINUX
    struct addrinfo *srcres, *dstres;
    struct ifaliasreq addreq;
    int s;

    s = socket(AF_INET, SOCK_DGRAM, 0);
    if(getaddrinfo(src, NULL, NULL, &srcres))
    {
        fprintf(stderr, "Error in source address");
        close(s);
        return -1;
    }
    if(getaddrinfo(dst, NULL, NULL, &dstres))
    {
        fprintf(stderr, "Error in destination address");
        close(s);
        return -1;
    }

    memset(&addreq, 0, sizeof(addreq));
    strcpy(addreq.ifra_name, name);
    memcpy(&addreq.ifra_addr, srcres->ai_addr, srcres->ai_addr->sa_len);

    memcpy(&addreq.ifra_dstaddr, dstres->ai_addr, dstres->ai_addr->sa_len);

    if (ioctl(s, SIOCSIFPHYADDR, &addreq) < 0)
    {
        close(s);
        return -1;
    }
    close(s);
    return 0;
#else
    struct ip_tunnel_parm p;
    struct ifreq ifr;
    int fd;
    int err;
    
    memset(&p, 0, sizeof(p));
    memset(&ifr, 0, sizeof(ifr));
    p.iph.version = 4;
    p.iph.ihl = 5;
    p.iph.protocol = IPPROTO_IPV6;
    p.iph.saddr = inet_addr(src);
    p.iph.daddr = inet_addr(dst);
    p.iph.ttl = 64;
    strncpy(p.name, name, IFNAMSIZ);

    strncpy(ifr.ifr_name, name, IFNAMSIZ);
    ifr.ifr_ifru.ifru_data = (void*)&p;
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    err = ioctl(fd, SIOCCHGTUNNEL, &ifr);
    close(fd);
    return err;
#endif
}

char *get_ip_from_interface(char *name)
{
    char addr[NI_MAXHOST];
    char *ret;
#ifndef LINUX
    struct ifaddrs *ifp;

    getifaddrs(&ifp);
    
    for(; ifp; ifp = ifp->ifa_next)
    {
        if(strncmp(ifp->ifa_name, name, IFNAMSIZ) == 0 && 
                ifp->ifa_addr->sa_family == AF_INET)
        {
            getnameinfo(ifp->ifa_addr, ifp->ifa_addr->sa_len, addr, NI_MAXHOST,
                    NULL, 0, NI_NUMERICHOST);
            ret = addr;
            return ret;
        }
    }
    freeifaddrs(ifp);
#else
    struct ifreq ifr;
    int fd;

    memset(&ifr, 0, sizeof(ifr));
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd < 0)
    {
        perror("socket");
        return NULL;
    }

    strncpy(ifr.ifr_name, name, IFNAMSIZ);
    if (ioctl(fd, SIOCGIFADDR, &ifr) == 0)
    {
        getnameinfo((struct sockaddr*)&ifr.ifr_addr, sizeof(struct in_addr), addr, NI_MAXHOST,
                NULL, 0, NI_NUMERICHOST);
        ret = addr;
        return ret;
    }
#endif
    return NULL;
}
